// jmode (XIM Server)
// connection management

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <list>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <unistd.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include "xim.h"
#include <map>

class sock_connection : public Connection{
public:
    sock_connection(int fd);
    int get_fd();
    int read_proc();
    int write_proc();
    bool write_pending();
    bool bConnect();
    void Close();
    void OnPushPacket();
private:
    void do_send_packet(TxPacket*);
    int m_fd;
    bool m_bConnection;
};

// global data for sock_connection
static int g_serv_fd;
int g_serv_port;

static map<int,sock_connection *> g_sock_connections;

//
static sock_connection *create_sock_connection(int fd)
{
    return new sock_connection(fd);
}

sock_connection::sock_connection(int fd)
{
    m_fd = fd;
    m_bConnection = true;
}

void sock_connection::Close()
{
    if ( m_bConnection ){
        close(m_fd);
        m_bConnection = false;
        remove_fd_watch(m_fd);
    }
}

void sock_connection::OnPushPacket()
{
    set_fd_watch_mask(m_fd,READ_OK|WRITE_OK);
}

int sock_connection::get_fd()
{
    return m_fd;
}

bool sock_connection::bConnect()
{
    return m_bConnection;
}

int sock_connection::read_proc()
{
    char h[5];
    char *buf;
    int l,m;
    if ( !m_bConnection ){
        return -1;
    }

    // ѥåȤΥإå major(1) minor(1) length(2)4Х
    // XIM_CONNECT1Х
    l = recv( m_fd,h,5 , MSG_PEEK);
    if ( l == 0){
        m_bConnection = false;
        return -1;
    }
    
    if ( mByteorder == BYTEORDER_UNKNOWN ){
        // Хȥξ֤XIM_CONNECT
        // Хȥꤹ롣
        if ( h[0] != XIM_CONNECT ){
            return -1;
        }
        if ( h[4] == 0x42 ){
            mByteorder = MSB_FIRST;
        }else if ( h[4] == 0x6c ){
            mByteorder = LSB_FIRST;
        }else{
            return -1;
        }
    }

    l = RxPacket::get_len((unsigned char *)h,mByteorder);        

    buf = (char *)malloc(l);
    m = recv(m_fd,buf,l,MSG_PEEK);
    if ( m == l ){
        recv(m_fd,buf,l,0); // discard
        mRxQ.push_back(createRxPacket((unsigned char *)buf,mByteorder));
        OnRecv();
    }
    free(buf);
    return 0;
}

int sock_connection::write_proc()
{
    list<TxPacket *>::iterator i;

    if ( !mTxQ.size()){
	return 0;
    }

    OnSend();
  
    while ( mPTxQ.size()){
	i = mPTxQ.begin();
	do_send_packet(*i);
	delete *i;
	mPTxQ.pop_front();
    }

    while (mTxQ.size()){
	i = mTxQ.begin();
	do_send_packet(*i);
	delete *i;
	mTxQ.pop_front();
    }
  
    if ( mIsCloseWait ){
	//񤭽Фλ
	m_bConnection = false;
	remove_fd_watch(m_fd);
	close(m_fd);
	return -1;
    }
    return 0;
}

void sock_connection::do_send_packet(TxPacket *tp)
{
    int buflen;
    char *buf;
    buflen = tp->get_length();
    buf = (char *)malloc (buflen);
    tp->write_to_buf((unsigned char *)buf,buflen,mByteorder);
    write(m_fd,buf,buflen);
    free(buf);
}

bool sock_connection::write_pending()
{
    if ( mTxQ.size()){
        return true;
    }
    return false;
}

void add_new_connection(int fd)
{
    pair<int,sock_connection *> p(fd,create_sock_connection(fd));
    g_sock_connections.insert(p);
    add_fd_watch(fd,READ_OK,sock_proc);

    struct linger l;
    l.l_onoff =1;
    l.l_linger = 100;
    setsockopt(fd,SOL_SOCKET,SO_LINGER, (SETSOCKOPT_TYPE_ARG4) &l,sizeof(l));
}

void accept_proc()
{
    int fd;
    struct sockaddr_in peer;
    socklen_t st = sizeof(struct sockaddr_in);

    fd = accept(g_serv_fd,(struct sockaddr *)&peer,&st);
    if ( fd != -1){
        add_new_connection(fd);
    }else{
        printf("failed to accept.\n");
    }
}

int setup_servfd()
{
    int foo;
    int serv_fd;
    struct sockaddr_in myhost;
    serv_fd = socket(AF_INET , SOCK_STREAM , 0);
    if ( serv_fd == -1 ){
        perror("failed in socket()");
        exit(1);
    }

    foo = 1;
    foo = setsockopt(serv_fd,SOL_SOCKET,SO_REUSEADDR,
		     (SETSOCKOPT_TYPE_ARG4) &foo, sizeof (foo));
    if ( foo == -1){
	perror("failed to set reuse option");
    }


    myhost.sin_family=AF_INET;
    myhost.sin_port=htons(0);//ݡȤϻꤻ
    myhost.sin_addr.s_addr=htonl(INADDR_ANY);
    foo = bind(serv_fd,(struct sockaddr *)&myhost,16);
    if(foo == -1){
        return -1;
    }

    // ݡֹ
    foo = sizeof(myhost);
    if ( getsockname(serv_fd,(sockaddr*)&myhost,(socklen_t*)&foo) ==-1){
	return -1;
    }
    
    g_serv_port = ntohs(myhost.sin_port);
    if ( g_option_mask & OPT_TRACE ){
	printf("TCP port %d is used for XIM connection.\n",g_serv_port);
    }

    foo = listen(serv_fd,5);
    if ( foo == -1){
	return -1;
    }
    return serv_fd;
}

void sock_proc(int fd,int ev)
{
    if ( fd == g_serv_fd ){
        accept_proc();
    }
    map<int,sock_connection *>::iterator i;
    i = g_sock_connections.find(fd);
    if ( i == g_sock_connections.end()){
        // XXX
        return ;
    }
    sock_connection *c;
    c = (*i).second;
    if ( ev & READ_OK ){
        //³ڤƤƤ⡢selectɤ߹߲ǽ֤Τǡˤ롣
        int r;
        r = c->read_proc();
        if ( r == -1){
            c->Close();
            g_sock_connections.erase(i);
            remove_fd_watch(fd);
            delete c;
	    if ( g_option_mask & OPT_TRACE ){
		printf("xim connection(sock). read error\n");
	    }
            return ;
        }
    }
    if ( ev & WRITE_OK ){
	if ( c->write_proc() < 0){
	    delete c;
	    g_sock_connections.erase(i);
	    if ( g_option_mask & OPT_TRACE ){
		printf("close xim connection(sock).\n");
	    }
	    return ;
	}
    }
    if ( !c->write_pending()){
        set_fd_watch_mask(fd,READ_OK);
    }
}

int sock_setup()
{
    return g_serv_fd = setup_servfd();
}
/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 */
